;; Generated package description from project.el  -*- no-byte-compile: t -*-
(define-package "project" "0.8.3" "Operations on the current project" '((emacs "26.1") (xref "1.4.0")) :commit "c0474ca21dd76cafbd1a04fd32d4a0bd42de6351" :url "https://elpa.gnu.org/packages/project.html")
